#include <stdexcept>
#include <msp/core/raii.h>
#include <msp/gl/extensions/arb_direct_state_access.h>
#include "bindable.h"
#include "buffer.h"
#include "bufferable.h"

using namespace std;

namespace Msp {
namespace GL {

Bufferable::Bufferable():
	buffer(0),
	offset(0),
	next_in_buffer(0),
	prev_in_buffer(0),
	location_dirty(false),
	dirty(false)
{ }

Bufferable::~Bufferable()
{
	unlink_from_buffer();
}

void Bufferable::use_buffer(Buffer *buf, Bufferable *prev)
{
	if(prev && buf!=prev->buffer)
		throw invalid_argument("Bufferable::use_buffer");

	if(buffer)
		unlink_from_buffer();

	buffer = buf;
	if(buffer)
	{
		prev_in_buffer = prev;
		if(prev_in_buffer)
		{
			next_in_buffer = prev_in_buffer->next_in_buffer;
			prev_in_buffer->next_in_buffer = this;
		}
	}

	location_dirty = true;
	dirty = true;
	update_offset();
}

void Bufferable::change_buffer(Buffer *buf)
{
	for(Bufferable *b=this; b; b=b->next_in_buffer)
	{
		b->buffer = buf;
		b->location_dirty = true;
		b->dirty = true;
	}
	for(Bufferable *b=prev_in_buffer; b; b=b->prev_in_buffer)
	{
		b->buffer = buf;
		b->location_dirty = true;
		b->dirty = true;
	}
}

unsigned Bufferable::get_required_buffer_size() const
{
	const Bufferable *last = this;
	for(; last->next_in_buffer; last=last->next_in_buffer) ;
	return last->offset+last->get_data_size();
}

Bufferable::AsyncUpdater *Bufferable::refresh_async() const
{
	return dirty ? new AsyncUpdater(*this) : 0;
}

void Bufferable::unlink_from_buffer()
{
	if(prev_in_buffer)
		prev_in_buffer->next_in_buffer = next_in_buffer;
	if(next_in_buffer)
	{
		next_in_buffer->prev_in_buffer = prev_in_buffer;
		next_in_buffer->update_offset();
	}
	prev_in_buffer = 0;
	next_in_buffer = 0;
	buffer = 0;
	offset = 0;
}

void Bufferable::update_offset()
{
	unsigned new_offset = 0;
	if(prev_in_buffer)
		new_offset = prev_in_buffer->offset+prev_in_buffer->get_data_size();

	unsigned align = get_alignment();
	new_offset += align-1;
	new_offset -= new_offset%align;
	if(new_offset!=offset)
	{
		offset = new_offset;
		location_dirty = true;
		dirty = true;
	}

	if(next_in_buffer)
		next_in_buffer->update_offset();
	else if(buffer && offset+get_data_size()>buffer->get_size())
	{
		location_dirty = true;
		dirty = true;
	}
}

void Bufferable::upload_data(char *target) const
{
	unsigned data_size = get_data_size();
	if(location_dirty)
	{
		buffer->require_size(offset+data_size);
		location_changed(buffer, offset, data_size);
		location_dirty = false;
	}

	if(target)
	{
		const char *source = reinterpret_cast<const char *>(get_data_pointer());
		copy(source, source+data_size, target);
	}
	else
		buffer->sub_data(offset, data_size, get_data_pointer());
	dirty = false;
}


Bufferable::AsyncUpdater::AsyncUpdater(const Bufferable &b):
	bufferable(b)
{
	bufferable.buffer->require_size(bufferable.get_required_buffer_size());
	mapped_address = reinterpret_cast<char *>(bufferable.buffer->map());
}

Bufferable::AsyncUpdater::~AsyncUpdater()
{
	bufferable.buffer->unmap();
}

void Bufferable::AsyncUpdater::upload_data()
{
	bufferable.upload_data(mapped_address+bufferable.offset);
	// Update all bufferables in the same buffer at once
	for(const Bufferable *b=bufferable.prev_in_buffer; b; b=b->prev_in_buffer)
		if(b->dirty)
			b->upload_data(mapped_address+b->offset);
	for(const Bufferable *b=bufferable.next_in_buffer; b; b=b->next_in_buffer)
		if(b->dirty)
			b->upload_data(mapped_address+b->offset);
}

} // namespace GL
} // namespace Msp
