#include <algorithm>
#include <cmath>
#include "environmentmap.h"
#include "renderer.h"
#include "resources.h"
#include "texunit.h"

using namespace std;

namespace Msp {
namespace GL {

EnvironmentMap::EnvironmentMap(Resources &resources, unsigned s, Renderable &r, Renderable &e):
	Effect(r),
	size(s),
	environment(e),
	sampler(resources.get<Sampler>("_linear_clamp.samp")),
	rendered(false),
	update_interval(1),
	update_delay(0)
{
	env_tex.storage(RGB8, size, 1);
	depth_buf.storage(DEPTH_COMPONENT32F, size, size);
	for(unsigned i=0; i<6; ++i)
	{
		fbo[i].attach(COLOR_ATTACHMENT0, env_tex, TextureCube::enumerate_faces(i), 0);
		fbo[i].attach(DEPTH_ATTACHMENT, depth_buf);
		fbo[i].require_complete();
	}

	camera.set_field_of_view(Geometry::Angle<float>::right());
	camera.set_aspect_ratio(1);
	camera.set_depth_clip(0.1, 100);
}

void EnvironmentMap::set_depth_clip(float n, float f)
{
	camera.set_depth_clip(n, f);
}

void EnvironmentMap::set_update_interval(unsigned i)
{
	update_interval = i;
	update_delay = min(update_delay, update_interval-1);
}

void EnvironmentMap::queue_update()
{
	update_delay = 0;
}

void EnvironmentMap::setup_frame(Renderer &renderer)
{
	if(rendered)
		return;

	rendered = true;
	renderable.setup_frame(renderer);

	if(update_delay)
	{
		if(update_interval)
			--update_delay;
		return;
	}
	update_delay = update_interval-1;
	environment.setup_frame(renderer);

	const Matrix *matrix = renderable.get_matrix();
	if(!matrix)
		return;

	Renderer::Push push(renderer);
	Renderer::Exclude exclude1(renderer, renderable);
	Renderer::Exclude exclude2(renderer, *this);

	camera.set_position(matrix->column(3).slice<3>(0));

	BindRestore bind_fbo(fbo[0]);
	for(unsigned i=0; i<6; ++i)
	{
		TextureCubeFace face = TextureCube::enumerate_faces(i);
		fbo[i].bind();
		fbo[i].clear();
		camera.set_look_direction(TextureCube::get_face_direction(face));
		camera.set_up_direction(TextureCube::get_t_direction(face));
		renderer.set_camera(camera);
		renderer.render(environment);
	}
}

void EnvironmentMap::finish_frame()
{
	if(rendered)
	{
		rendered = false;
		renderable.finish_frame();
		environment.finish_frame();
	}
}

void EnvironmentMap::render(Renderer &renderer, Tag tag) const
{
	if(!enabled_passes.count(tag))
		return renderer.render(renderable, tag);

	Renderer::Push _push_rend(renderer);

	unsigned unit = renderer.allocate_effect_texunit();
	shdata.uniform("environment_map", static_cast<int>(unit));
	Bind _bind_sampler(sampler, unit);
	Bind _bind_env(env_tex, unit);

	const Matrix &camera_matrix = renderer.get_camera()->get_object_matrix();
	shdata.uniform("env_eye_matrix", camera_matrix.block<3, 3>(0, 0));

	renderer.add_shader_data(shdata);
	renderer.render(renderable, tag);
}

} // namespace GL
} // namespace Msp
